# ==================================================================================================
# Copyright 2012 Twitter, Inc.
# --------------------------------------------------------------------------------------------------
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this work except in compliance with the License.
# You may obtain a copy of the License in the LICENSE file, or at:
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==================================================================================================

from contextlib import contextmanager
import os
import sys
import threading
import time


__all__ = ('Tracer',)


class Trace(object):
  __slots__ = ('msg', 'verbosity', 'parent', 'children', '_clock', '_start', '_stop')
  def __init__(self, msg, parent=None, verbosity=1, clock=time):
    self.msg = msg
    self.verbosity = verbosity
    self.parent = parent
    if parent is not None:
      parent.children.append(self)
    self.children = []
    self._clock = clock
    self._start = self._clock.time()
    self._stop = None

  def stop(self):
    self._stop = self._clock.time()

  def duration(self):
    assert self._stop is not None
    return self._stop - self._start


class Tracer(object):
  """
    A multi-threaded tracer.
  """
  @classmethod
  def env_filter(cls, env_variable):
    def predicate(verbosity):
      try:
        env_verbosity = int(os.environ.get(env_variable, -1))
      except ValueError:
        env_verbosity = -1
      return verbosity <= env_verbosity
    return predicate

  def __init__(self, predicate=None, output=sys.stderr, clock=time):
    """
      If predicate specified, it should take a "verbosity" integer and determine whether
      or not to log, e.g.

        def predicate(verbosity):
          try:
            return verbosity < int(os.environ.get('APP_VERBOSITY', 0))
          except ValueError:
            return False

      output defaults to sys.stderr, but can take any file-like object.
    """
    self._predicate = predicate or (lambda verbosity: True)
    self._length = None
    self._output = output
    self._isatty = getattr(output, 'isatty', False) and output.isatty()
    self._lock = threading.RLock()
    self._local = threading.local()
    self._clock = clock

  def should_log(self, V):
    return self._predicate(V)

  def log(self, msg, V=0, end='\n'):
    if not self.should_log(V):
      return
    if not self._isatty and end == '\r':
      # force newlines if we're not a tty
      end = '\n'
    trailing_whitespace = ''
    with self._lock:
      if self._length and self._length > len(msg):
        trailing_whitespace = ' ' * (self._length - len(msg))
      self._output.write(msg + trailing_whitespace + end)
      self._output.flush()
      self._length = len(msg) if end == '\r' else 0

  def print_trace_snippet(self):
    parent = self._local.parent
    parent_verbosity = parent.verbosity
    if not self.should_log(parent_verbosity):
      return
    traces = []
    while parent:
      if self.should_log(parent.verbosity):
        traces.append(parent.msg)
      parent = parent.parent
    self.log(' :: '.join(reversed(traces)), V=parent_verbosity, end='\r')

  def print_trace(self, indent=0, node=None):
    node = node or self._local.parent
    with self._lock:
      self.log(' ' * indent + ('%s: %.1fms' % (node.msg, 1000.0 * node.duration())),
               V=node.verbosity)
      for child in node.children:
        self.print_trace(indent=indent + 2, node=child)

  @contextmanager
  def timed(self, msg, V=0):
    if getattr(self._local, 'parent', None) is None:
      self._local.parent = Trace(msg, verbosity=V, clock=self._clock)
    else:
      parent = self._local.parent
      self._local.parent = Trace(msg, parent=parent, verbosity=V, clock=self._clock)
    self.print_trace_snippet()
    yield
    self._local.parent.stop()
    if self._local.parent.parent is not None:
      self._local.parent = self._local.parent.parent
    else:
      self.print_trace()
      self._local.parent = None


def main(args):
  import random

  tracer = Tracer(output=open(args[0], 'w')) if len(args) > 0 else Tracer()

  def process(name):
    with tracer.timed(name):
      with tracer.timed('acquiring'):
        with tracer.timed('downloading'):
          time.sleep(3 * random.random())
          if random.random() > 0.66:
            tracer.log('%s failed downloading!' % name)
            return
        with tracer.timed('unpacking'):
          time.sleep(1 * random.random())
      with tracer.timed('configuring'):
        time.sleep(0.5 * random.random())
      with tracer.timed('building'):
        time.sleep(5.0 * random.random())
        if random.random() > 0.66:
          tracer.log('%s failed building!' % name)
          return
      with tracer.timed('installing'):
        time.sleep(2.0 * random.random())

  workers = [threading.Thread(target=process, args=('worker %d' % k,)) for k in range(5)]
  for worker in workers:
    worker.start()
  for worker in workers:
    worker.join()


if __name__ == '__main__':
  main(sys.argv[1:])



from datetime import datetime, timedelta

from twitter.common.lang import total_ordering

# TODO(wickman) Do something that won't break if this is running over NYE?
_CURRENT_YEAR = str(datetime.now().year)


class Level(object):
  DEBUG   = 0
  INFO    = 10
  WARNING = 20
  ERROR   = 30
  FATAL   = 40


@total_ordering
class Line(object):
  __slots__ = (
    'raw',
    'level',
    'datetime',
    'pid',
    'source',
    'message'
  )

  @classmethod
  def parse(cls, line):
    """parses a line and returns Line if successfully parsed, ValueError/None otherwise."""
    raise NotImplementedError

  @staticmethod
  def parse_order(line, *line_parsers):
    """Given a text line and any number of Line implementations, return the first that matches
       or None if no lines match."""
    for parser in line_parsers:
      try:
        return parser.parse(line)
      except ValueError:
        continue

  def __init__(self, raw, level, dt, pid, source, message):
    (self.raw, self.level, self.datetime, self.pid, self.source, self.message) = (
        raw, level, dt, pid, source, message)

  def extend(self, lines):
    extension = '\n'.join(lines)
    return self.__class__('\n'.join([self.raw, extension]), self.level, self.datetime, self.pid,
        self.source, '\n'.join([self.message, extension]))

  def __lt__(self, other):
    return self.datetime < other.datetime

  def __gt__(self, other):
    return self.datetime > other.datetime

  def __eq__(self, other):
    if not isinstance(other, self.__class__):
      return False
    return (self.datetime == other.datetime and
            self.level == other.level and
            self.pid == other.pid and
            self.source == other.source and
            self.message == other.message)

  def __str__(self):
    return self.raw


class GlogLine(Line):
  LEVEL_MAP = {
    'I': Level.INFO,
    'W': Level.WARNING,
    'E': Level.ERROR,
    'F': Level.FATAL,
    'D': Level.DEBUG
  }

  @classmethod
  def split_time(cls, line):
    if len(line) == 0:
      raise ValueError
    if line[0] not in 'IWEFD':
      raise ValueError
    sline = line[1:].split(' ')
    if len(sline) < 2:
      raise ValueError
    t = datetime.strptime(''.join([_CURRENT_YEAR, sline[0], ' ', sline[1]]), '%Y%m%d %H:%M:%S.%f')
    return cls.LEVEL_MAP[line[0]], t, sline[2:]

  @classmethod
  def parse(cls, line):
    level, dt, rest = cls.split_time(line)
    pid, source, message = rest[0], rest[1], ' '.join(rest[2:])
    return cls(line, level, dt, pid, source, message)


class ZooLine(Line):
  LEVEL_MAP = {
    "ZOO_INVALID": 0,
    "ZOO_ERROR": Level.ERROR,
    "ZOO_WARN": Level.WARNING,
    "ZOO_INFO": Level.INFO,
    "ZOO_DEBUG": Level.DEBUG
  }

  @classmethod
  def parse(cls, line):
    sline = line.split(':')
    if len(sline) < 6:
      raise ValueError
    t = datetime.strptime(':'.join(sline[0:3]), '%Y-%m-%d %H:%M:%S,%f')
    pid = sline[3]
    ssource = sline[4].split('@')
    level = cls.LEVEL_MAP.get(ssource[0], 0)
    source = '@'.join(ssource[1:])
    return cls(line, level, t, pid, source, ':'.join(sline[5:]))



# ==================================================================================================
# Copyright 2011 Twitter, Inc.
# --------------------------------------------------------------------------------------------------
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this work except in compliance with the License.
# You may obtain a copy of the License in the LICENSE file, or at:
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==================================================================================================

"""
Glog log system global options.

Exports module-level options such as --log_dir and --stderr_log_level, but may be
overridden locally before calling log.init().
"""

from __future__ import print_function

import logging
import optparse
import sys


_DISK_LOG_LEVEL_OPTION = 'twitter_common_log_disk_log_level'


_DEFAULT_LOG_OPTS = {
  'twitter_common_log_stderr_log_level': 'ERROR',
  _DISK_LOG_LEVEL_OPTION: 'INFO',
  'twitter_common_log_log_dir': '/var/tmp',
  'twitter_common_log_simple': False,
  'twitter_common_log_scribe_buffer': False,
  'twitter_common_log_scribe_host': 'localhost',
  'twitter_common_log_scribe_log_level': 'NONE',
  'twitter_common_log_scribe_port': 1463,
  'twitter_common_log_scribe_category': 'python_default'
}


try:
  from twitter.common import app
  HAVE_APP = True
except ImportError:
  from collections import namedtuple
  DefaultLogOpts = namedtuple('DefaultLogOpts', _DEFAULT_LOG_OPTS.keys())
  class AppDefaultProxy(object):
    def __init__(self):
      self._opts = DefaultLogOpts(**_DEFAULT_LOG_OPTS)
    def get_options(self):
      return self._opts
    def set_option(self, key, value, force=False):
      opts = self._opts._asdict()
      if force or key not in opts:
        opts[key] = value
  app = AppDefaultProxy()
  HAVE_APP = False


class LogOptionsException(Exception): pass


class LogOptions(object):
  _LOG_LEVEL_NONE_KEY = 'NONE'
  LOG_LEVEL_NONE = 100

  _LOG_LEVELS = {
    'DEBUG':             logging.DEBUG,
    'INFO':              logging.INFO,
    'WARN':              logging.WARN,
    'FATAL':             logging.FATAL,
    'ERROR':             logging.ERROR,
    _LOG_LEVEL_NONE_KEY: LOG_LEVEL_NONE
  }

  _LOG_SCHEMES = [
    'google',
    'plain'
  ]

  _STDERR_LOG_LEVEL = None
  _STDOUT_LOG_SCHEME = None
  _DISK_LOG_LEVEL = None
  _DISK_LOG_SCHEME = None
  _LOG_DIR = None
  _SIMPLE = None
  _SCRIBE_BUFFER = None
  _SCRIBE_HOST = None
  _SCRIBE_LOG_LEVEL = None
  _SCRIBE_LOG_SCHEME = None
  _SCRIBE_PORT = None
  _SCRIBE_CATEGORY = None

  @staticmethod
  def _parse_loglevel(log_level, scheme='google'):
    level = None
    components = log_level.split(':')
    if len(components) == 1:
      level = components[0]
    elif len(components) == 2:
      scheme, level = components[0], components[1]
    else:
      raise LogOptionsException("Malformed log level: %s" % log_level)

    if level in LogOptions._LOG_LEVELS:
      level = LogOptions._LOG_LEVELS[level]
    else:
      raise LogOptionsException("Unknown log level: %s" % level)

    if scheme not in LogOptions._LOG_SCHEMES:
      raise LogOptionsException("Unknown log scheme: %s" % scheme)

    return (scheme, level)

  @staticmethod
  def loglevel_name(log_level):
    """
      Return the log level name of the given log_level (integer), or None if it has no name.
    """
    for name, value in LogOptions._LOG_LEVELS.items():
      if value == log_level:
        return name

  @staticmethod
  def _valid_loglevel(log_level):
    try:
      LogOptions._parse_loglevel(log_level)
      return True
    except:
      return False

  @staticmethod
  def _is_scribe_logging_required():
    return LogOptions._LOG_LEVEL_NONE_KEY != app.get_options().twitter_common_log_scribe_log_level

  @staticmethod
  def disable_scribe_logging():
    """
      Disables scribe logging altogether.
    """
    app.set_option("_SCRIBE_LOG_LEVEL", LogOptions._LOG_LEVEL_NONE_KEY, force=True)

  @staticmethod
  def set_scribe_category(category):
    """
      Set the scribe category for logging. Must be called before log.init() for
      changes to take effect.
    """
    LogOptions._SCRIBE_CATEGORY = category

  @staticmethod
  def scribe_category():
    """
      Get the current category used when logging to the scribe daemon.
    """
    if LogOptions._SCRIBE_CATEGORY is None:
      LogOptions._SCRIBE_CATEGORY = app.get_options().twitter_common_log_scribe_category
    return LogOptions._SCRIBE_CATEGORY

  @staticmethod
  def set_scribe_log_level(log_level):
    """
      Set the log level for scribe.
    """
    LogOptions._SCRIBE_LOG_SCHEME, LogOptions._SCRIBE_LOG_LEVEL = (
      LogOptions._parse_loglevel(log_level, scheme='google')
    )

  @staticmethod
  def scribe_log_level():
    """
      Get the current scribe_log_level (in logging units specified by logging module.)
    """
    if LogOptions._SCRIBE_LOG_LEVEL is None:
      LogOptions.set_scribe_log_level(app.get_options().twitter_common_log_scribe_log_level)
    return LogOptions._SCRIBE_LOG_LEVEL

  @staticmethod
  def scribe_log_scheme():
    """
      Get the current scribe log scheme.
    """
    if LogOptions._SCRIBE_LOG_SCHEME is None:
      LogOptions.set_scribe_log_level(app.get_options().twitter_common_log_scribe_log_level)
    return LogOptions._SCRIBE_LOG_SCHEME

  @staticmethod
  def set_scribe_buffer(buffer_enabled):
    """
      Set buffer for scribe logging. Must be called before log.init() for
      changes to take effect.
    """
    LogOptions._SCRIBE_BUFFER = buffer_enabled

  @staticmethod
  def scribe_buffer():
    """
      Get the current buffer setting for scribe logging.
    """
    if LogOptions._SCRIBE_BUFFER is None:
      LogOptions._SCRIBE_BUFFER = app.get_options().twitter_common_log_scribe_buffer
    return LogOptions._SCRIBE_BUFFER

  @staticmethod
  def set_scribe_host(host):
    """
      Set the scribe host for logging. Must be called before log.init() for
      changes to take effect.
    """
    LogOptions._SCRIBE_HOST = host

  @staticmethod
  def scribe_host():
    """
      Get the current host running the scribe daemon.
    """
    if LogOptions._SCRIBE_HOST is None:
      LogOptions._SCRIBE_HOST = app.get_options().twitter_common_log_scribe_host
    return LogOptions._SCRIBE_HOST

  @staticmethod
  def set_scribe_port(port):
    """
      Set the scribe port for logging. Must be called before log.init() for
      changes to take effect.
    """
    LogOptions._SCRIBE_PORT = port

  @staticmethod
  def scribe_port():
    """
      Get the current port used to connect to the scribe daemon.
    """
    if LogOptions._SCRIBE_PORT is None:
      LogOptions._SCRIBE_PORT = app.get_options().twitter_common_log_scribe_port
    return LogOptions._SCRIBE_PORT

  @staticmethod
  def set_stderr_log_level(log_level):
    """
      Set the log level for stderr.
    """
    LogOptions._STDOUT_LOG_SCHEME, LogOptions._STDERR_LOG_LEVEL = (
        LogOptions._parse_loglevel(log_level, scheme='plain'))

  @staticmethod
  def stderr_log_level():
    """
      Get the current stderr_log_level (in logging units specified by logging module.)
    """
    if LogOptions._STDERR_LOG_LEVEL is None:
      LogOptions.set_stderr_log_level(app.get_options().twitter_common_log_stderr_log_level)
    return LogOptions._STDERR_LOG_LEVEL

  @staticmethod
  def stderr_log_scheme():
    """
      Get the current stderr log scheme.
    """
    if LogOptions._STDOUT_LOG_SCHEME is None:
      LogOptions.set_stderr_log_level(app.get_options().twitter_common_log_stderr_log_level)
    return LogOptions._STDOUT_LOG_SCHEME

  # old deprecated version of these functions.
  set_stdout_log_level = set_stderr_log_level
  stdout_log_level = stderr_log_level
  stdout_log_scheme = stderr_log_scheme

  @staticmethod
  def _is_disk_logging_required():
    return LogOptions._LOG_LEVEL_NONE_KEY != getattr(app.get_options(), _DISK_LOG_LEVEL_OPTION)

  @staticmethod
  def disable_disk_logging():
    """
      Disables disk logging altogether.
    """
    app.set_option(_DISK_LOG_LEVEL_OPTION, LogOptions._LOG_LEVEL_NONE_KEY, force=True)

  @staticmethod
  def set_disk_log_level(log_level):
    """
      Set the log level for disk.
    """
    LogOptions._DISK_LOG_SCHEME, LogOptions._DISK_LOG_LEVEL = (
        LogOptions._parse_loglevel(log_level, scheme='google'))

  @staticmethod
  def disk_log_level():
    """
      Get the current disk_log_level (in logging units specified by logging module.)
    """
    if LogOptions._DISK_LOG_LEVEL is None:
      LogOptions.set_disk_log_level(app.get_options().twitter_common_log_disk_log_level)
    return LogOptions._DISK_LOG_LEVEL

  @staticmethod
  def disk_log_scheme():
    """
      Get the current disk log scheme.
    """
    if LogOptions._DISK_LOG_SCHEME is None:
      LogOptions.set_disk_log_level(app.get_options().twitter_common_log_disk_log_level)
    return LogOptions._DISK_LOG_SCHEME

  @staticmethod
  def set_log_dir(dir):
    """
      Set the logging dir for disk logging.  Must be called before log.init() for
      changes to take effect.
    """
    LogOptions._LOG_DIR = dir

  @staticmethod
  def log_dir():
    """
      Get the current directory into which logs will be written.
    """
    if LogOptions._LOG_DIR is None:
      LogOptions._LOG_DIR = app.get_options().twitter_common_log_log_dir
    return LogOptions._LOG_DIR

  @staticmethod
  def set_simple(value):
    """
      Enable/disable simple logging mode.  Must be called before log.init().
    """
    LogOptions._SIMPLE = bool(value)

  @staticmethod
  def simple():
    """
      Whether or not simple logging should be used.
    """
    if LogOptions._SIMPLE is None:
      LogOptions._SIMPLE = app.get_options().twitter_common_log_simple
    return LogOptions._SIMPLE

  @staticmethod
  def _disk_options_callback(option, opt, value, parser):
    try:
      LogOptions.set_disk_log_level(value)
    except LogOptionsException as e:
      raise optparse.OptionValueError('Failed to parse option: %s' % e)
    parser.values.twitter_common_log_disk_log_level = value

  @staticmethod
  def _scribe_options_callback(option, opt, value, parser):
    try:
      LogOptions.set_scribe_log_level(value)
    except LogOptionsException as e:
      raise optparse.OptionValueError('Failed to parse option: %s' % e)
    parser.values.twitter_common_log_scribe_log_level = value

  __log_to_stderr_is_set = False
  __log_to_stdout_is_set = False
  @staticmethod
  def _stderr_options_callback(option, opt, value, parser):
    if LogOptions.__log_to_stdout_is_set:
      raise optparse.OptionValueError(
        '--log_to_stdout is an obsolete flag that was replaced by --log_to_stderr. '
        'Use only --log_to_stderr.')
    LogOptions.__log_to_stderr_is_set = True
    try:
      LogOptions.set_stderr_log_level(value)
    except LogOptionsException as e:
      raise optparse.OptionValueError('Failed to parse option: %s' % e)
    parser.values.twitter_common_log_stderr_log_level = value

  @staticmethod
  def _stdout_options_callback(option, opt, value, parser):
    if LogOptions.__log_to_stderr_is_set:
      raise optparse.OptionValueError(
        '--log_to_stdout is an obsolete flag that was replaced by --log_to_stderr. '
        'Use only --log_to_stderr.')
    LogOptions.__log_to_stdout_is_set = True
    print('--log_to_stdout is an obsolete flag that was replaced by '
          '--log_to_stderr. Use --log_to_stderr instead.', file=sys.stderr)
    try:
      LogOptions.set_stderr_log_level(value)
    except LogOptionsException as e:
      raise optparse.OptionValueError('Failed to parse option: %s' % e)
    parser.values.twitter_common_log_stderr_log_level = value


_LOGGING_HELP = """The level at which logging to %%s [default: %%%%default].
Takes either LEVEL or scheme:LEVEL, where LEVEL is one
of %s and scheme is one of %s.
""" % (repr(LogOptions._LOG_LEVELS.keys()), repr(LogOptions._LOG_SCHEMES))


if HAVE_APP:
  app.add_option('--log_to_stdout',
                 callback=LogOptions._stdout_options_callback,
                 default=_DEFAULT_LOG_OPTS['twitter_common_log_stderr_log_level'],
                 type='string',
                 action='callback',
                 metavar='[scheme:]LEVEL',
                 dest='twitter_common_log_stderr_log_level',
                 help='OBSOLETE - legacy flag, use --log_to_stderr instead.')

  app.add_option('--log_to_stderr',
                 callback=LogOptions._stderr_options_callback,
                 default=_DEFAULT_LOG_OPTS['twitter_common_log_stderr_log_level'],
                 type='string',
                 action='callback',
                 metavar='[scheme:]LEVEL',
                 dest='twitter_common_log_stderr_log_level',
                 help=_LOGGING_HELP % 'stderr')

  app.add_option('--log_to_disk',
                 callback=LogOptions._disk_options_callback,
                 default=_DEFAULT_LOG_OPTS['twitter_common_log_disk_log_level'],
                 type='string',
                 action='callback',
                 metavar='[scheme:]LEVEL',
                 dest='twitter_common_log_disk_log_level',
                 help=_LOGGING_HELP % 'disk')

  app.add_option('--log_dir',
                 type='string',
                 default=_DEFAULT_LOG_OPTS['twitter_common_log_log_dir'],
                 metavar='DIR',
                 dest='twitter_common_log_log_dir',
                 help='The directory into which log files will be generated [default: %default].')

  app.add_option('--log_simple',
                 default=_DEFAULT_LOG_OPTS['twitter_common_log_simple'],
                 action='store_true',
                 dest='twitter_common_log_simple',
                 help='Write a single log file rather than one log file per log level '
                      '[default: %default].')

  app.add_option('--log_to_scribe',
              callback=LogOptions._scribe_options_callback,
              default=_DEFAULT_LOG_OPTS['twitter_common_log_scribe_log_level'],
              type='string',
              action='callback',
              metavar='[scheme:]LEVEL',
              dest='twitter_common_log_scribe_log_level',
              help=_LOGGING_HELP % 'scribe')

  app.add_option('--scribe_category',
              type='string',
              default=_DEFAULT_LOG_OPTS['twitter_common_log_scribe_category'],
              metavar='CATEGORY',
              dest='twitter_common_log_scribe_category',
              help="The category used when logging to the scribe daemon. [default: %default].")

  app.add_option('--scribe_buffer',
              action='store_true',
              default=_DEFAULT_LOG_OPTS['twitter_common_log_scribe_buffer'],
              dest='twitter_common_log_scribe_buffer',
              help="Buffer messages when scribe is unavailable rather than dropping them. [default: %default].")

  app.add_option('--scribe_host',
              type='string',
              default=_DEFAULT_LOG_OPTS['twitter_common_log_scribe_host'],
              metavar='HOST',
              dest='twitter_common_log_scribe_host',
              help="The host running the scribe daemon. [default: %default].")

  app.add_option('--scribe_port',
              type='int',
              default=_DEFAULT_LOG_OPTS['twitter_common_log_scribe_port'],
              metavar='PORT',
              dest='twitter_common_log_scribe_port',
              help="The port used to connect to the scribe daemon. [default: %default].")



# ==================================================================================================
# Copyright 2013 Twitter, Inc.
# --------------------------------------------------------------------------------------------------
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this work except in compliance with the License.
# You may obtain a copy of the License in the LICENSE file, or at:
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==================================================================================================

from logging import Handler

try:
  from scribe import scribe
  from thrift.protocol import TBinaryProtocol
  from thrift.transport import TTransport, TSocket
  _SCRIBE_PRESENT = True
except ImportError:
  _SCRIBE_PRESENT = False


class ScribeHandler(Handler):
  """logging.Handler interface for Scribe."""
  class ScribeHandlerException(Exception):
    pass

  def __init__(self, *args, **kwargs):
    """logging.Handler interface for Scribe.

    Params:
    buffer: If True, buffer messages when scribe is unavailable. If False, drop on floor.
    category: Scribe category for logging events.
    host: Scribe host.
    port: Scribe port.
    """
    if not _SCRIBE_PRESENT:
      raise self.ScribeHandlerException(
        "Could not initialize ScribeHandler: Scribe modules not present.")
    self._buffer_enabled = kwargs.pop("buffer")
    self._category = kwargs.pop("category")
    self._client = None
    self._host = kwargs.pop("host")
    self._log_buffer = []
    self._port = kwargs.pop("port")
    self._transport = None
    Handler.__init__(self, *args, **kwargs)

  @property
  def messages_pending(self):
    """Return True if there are messages in the buffer."""
    return bool(self._log_buffer)

  @property
  def client(self):
    """Scribe client object."""
    if not self._client:
      protocol = TBinaryProtocol.TBinaryProtocol(trans=self.transport,
                                                 strictRead=False,
                                                 strictWrite=False)
      self._client = scribe.Client(iprot=protocol, oprot=protocol)
    return self._client

  @property
  def transport(self):
    """Scribe transport object."""
    if not self._transport:
      socket = TSocket.TSocket(host=self._host, port=self._port)
      self._transport = TTransport.TFramedTransport(socket)
    return self._transport

  def close(self):
    """Flushes any remaining messages in the queue."""
    if self.messages_pending:
      try:
        self.scribe_write(self._log_buffer)
      except self.ScribeHandlerException:
        pass
    Handler.close(self)

  def emit(self, record):
    """Emit a record via Scribe."""
    fmt_record = self.format(record)
    self._log_buffer.append(scribe.LogEntry(category=self._category, message=fmt_record))
    try:
      self.scribe_write(self._log_buffer)
    except self.ScribeHandlerException:
      if not self._buffer_enabled:
        self._log_buffer = []
    else:
      self._log_buffer = []

  def scribe_write(self, messages):
    """Sends a list of messages to scribe.

    Params:
    messages: List of scribe.LogEntry objects.

    Raises:
    ScribeHandlerException on timeouts and connection errors.
    """
    try:
      self.transport.open()
      result = self.client.Log(messages)
      if result != scribe.ResultCode.OK:
        raise self.ScribeHandlerException('Scribe message submission failed')
    except TTransport.TTransportException as err:
      raise self.ScribeHandlerException('Could not connect to scribe host=%s:%s error=%s'
                                        % (self._host, self._port, err))
    finally:
      self.transport.close()



# ==================================================================================================
# Copyright 2011 Twitter, Inc.
# --------------------------------------------------------------------------------------------------
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this work except in compliance with the License.
# You may obtain a copy of the License in the LICENSE file, or at:
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==================================================================================================

import logging
import socket
import sys
import time

from twitter.common.log.formatters.base import format_message


class GlogFormatter(logging.Formatter):
  """
    Format a log in Google style format:
    [DIWEF]mmdd hh:mm:ss.uuuuuu pid file:line] msg
  """
  SCHEME = 'google'
  LOG_LINE_FORMAT = "[DIWEF]mmdd hh:mm:ss.uuuuuu pid file:line] msg"

  LEVEL_MAP = {
    logging.FATAL: 'F',
    logging.ERROR: 'E',
    logging.WARN:  'W',
    logging.INFO:  'I',
    logging.DEBUG: 'D'
  }

  @classmethod
  def logfile_preamble(cls):
    return ''.join('%s\n' % line for line in [
      'Log file created at: %s' % time.strftime('%Y/%m/%d %H:%M:%S', time.localtime()),
      'Running on machine: %s' % socket.gethostname(),
      cls.LOG_LINE_FORMAT,
      'Command line: %s' % ' '.join(sys.argv)])

  def __init__(self):
    logging.Formatter.__init__(self)

  def format(self, record):
    try:
      level = GlogFormatter.LEVEL_MAP[record.levelno]
    except:
      level = '?'
    date = time.localtime(record.created)
    date_usec = (record.created - int(record.created)) * 1e6
    record_message = '%c%02d%02d %02d:%02d:%02d.%06d %s %s:%d] %s' % (
       level, date.tm_mon, date.tm_mday, date.tm_hour, date.tm_min, date.tm_sec, date_usec,
       record.process if record.process is not None else '?????',
       record.filename,
       record.lineno,
       format_message(record))
    record.getMessage = lambda: record_message
    return logging.Formatter.format(self, record)



# ==================================================================================================
# Copyright 2011 Twitter, Inc.
# --------------------------------------------------------------------------------------------------
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this work except in compliance with the License.
# You may obtain a copy of the License in the LICENSE file, or at:
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==================================================================================================

import logging
from twitter.common.log.formatters.base import format_message

class PlainFormatter(logging.Formatter):
  """
    Format a log in a plainer style:
    type] msg
  """
  SCHEME = 'plain'

  LEVEL_MAP = {
    logging.FATAL: 'FATAL',
    logging.ERROR: 'ERROR',
    logging.WARN:  ' WARN',
    logging.INFO:  ' INFO',
    logging.DEBUG: 'DEBUG'
  }

  def __init__(self):
    logging.Formatter.__init__(self)

  def format(self, record):
    try:
      level = PlainFormatter.LEVEL_MAP[record.levelno]
    except:
      level = '?????'
    record_message = '%s] %s' % (level, format_message(record))
    record.getMessage = lambda: record_message
    return logging.Formatter.format(self, record)



from collections import deque
from datetime import datetime, timedelta
import errno
from io import BytesIO, FileIO
import os

from twitter.common.lang import Compatibility

from .parsers import Line


class Buffer(object):
  CHUNKSIZE = 65536

  @classmethod
  def maybe_filelike(cls, filename_or_filelike):
    if isinstance(filename_or_filelike, Compatibility.string):
      return FileIO(filename_or_filelike)
    else:
      return filename_or_filelike

  @classmethod
  def reset(cls, fp):
    try:
      fp.seek(0, os.SEEK_CUR)
    except IOError:
      return False
    except OSError as e:
      if e.errno == errno.ESPIPE:
        return False
      raise

  def __init__(self, filename_or_filelike, infinite=False):
    self._fp = self.maybe_filelike(filename_or_filelike)
    self._buffer = deque()
    self._tail = None
    self._infinite = infinite

  def next(self):
    last_chunk = self.CHUNKSIZE
    while len(self._buffer) == 0 and last_chunk == self.CHUNKSIZE:
      tail_add = self._fp.read(self.CHUNKSIZE)
      if tail_add:
        self._tail = self._tail + tail_add if self._tail else tail_add
        self._flatten_tail()
      last_chunk = len(tail_add)
    if self._buffer:
      return self._buffer.popleft()
    if last_chunk != self.CHUNKSIZE and self._infinite:
      self.reset(self._fp)
    if last_chunk != self.CHUNKSIZE and self._tail is not None and not self._infinite:
      rv = self._tail
      self._tail = None
      return rv

  def _flatten_tail(self):
    flattened = self._tail.split('\n')
    if len(flattened) > 1:
      self._buffer.extend(flattened[0:-1])
      self._tail = flattened[-1]


class Stream(object):
  class EOF(object): pass

  def __init__(self, filename_or_filelike, parsers, infinite=False):
    """
      Given a filelike-object and a set of Line-derived parsers (e.g. GlogLine, ZooLine),
      generate Line objects from the stream.  If infinite=True, continue after hitting
      EOF.
    """
    self._buffer = Buffer(filename_or_filelike, infinite)
    self._head = None
    self._tail = []
    self._infinite = infinite
    self._parsers = parsers

  def _full_head(self):
    return self._head.extend(self._tail) if self._tail else self._head

  def _handle_end(self):
    if self._infinite:
      return
    else:
      if self._head:
        rv = self._full_head()
        self._head, self._tail = None, []
        return rv
      else:
        return self.EOF

  def next(self):
    while True:
      line = self._buffer.next()
      if line is None:
        return self._handle_end()
      tail = Line.parse_order(line, *self._parsers)
      if tail is not None:
        if self._head:
          rv = self._full_head()
          self._head, self._tail = tail, []
          return rv
        else:
          self._head, self._tail = tail, []
      else:
        self._tail.append(line)


class StreamMuxer(object):
  """
    Multiplexes a set of streams into a single stream.
  """
  def __init__(self, streams):
    """
      Takes a set of (stream, label) pairs.
    """
    streams = list(streams)
    self._labels = dict(streams)
    self._refresh = set(stream for (stream, _) in streams)
    self._heads = set()

  def _collect(self):
    discard = set()
    for stream in self._refresh:
      line = stream.next()
      if line is Stream.EOF:
        discard.add(stream)
      elif line is not None:
        discard.add(stream)
        self._heads.add((line, stream))
    self._refresh -= discard

  def _pop(self):
    try:
      minimum = min(self._heads)
      self._heads.discard(minimum)
      return minimum
    except ValueError:
      pass

  def next(self):
    """
      Returns (label, Line) pairs as they're available, None if nothing is available or Stream.EOF
      if all streams have terminated.
    """
    self._collect()
    if not self._heads and not self._refresh:
      return Stream.EOF
    minimum = self._pop()
    if minimum:
      line, stream = minimum
      self._refresh.add(stream)
      return (self._labels[stream], line)



# ==================================================================================================
# Copyright 2011 Twitter, Inc.
# --------------------------------------------------------------------------------------------------
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this work except in compliance with the License.
# You may obtain a copy of the License in the LICENSE file, or at:
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==================================================================================================

# TODO Backend support needs to be broken out into modules, a'la twitter.common.app.
__author__ = 'Brian Wickman'

import logging

from .initialize import (
  init,
  teardown_disk_logging,
  teardown_stderr_logging)
from .tracer import Tracer

try:
  from twitter.common import app
  from twitter.common.log.options import LogOptions

  class LoggingSubsystem(app.Module):
    def __init__(self):
      app.Module.__init__(self, __name__, description="Logging subsystem.")
    def setup_function(self):
      if not LogOptions._is_disk_logging_required():
        init()
      else:
        init(app.name())

  app.register_module(LoggingSubsystem())
except ImportError:
  # Do not require twitter.common.app
  pass

debug = logging.debug
info = logging.info
warning = logging.warning
warn = logging.warning
error = logging.error
fatal = logging.fatal
log = logging.log
logger = logging.getLogger

DEBUG = logging.DEBUG
INFO = logging.INFO
WARNING = logging.WARNING
WARN = logging.WARN
ERROR = logging.ERROR
FATAL = logging.FATAL

__all__ = (
  # directives
  'debug',
  'info',
  'warning',
  'warn', # alias
  'error',
  'fatal',
  'log',
  'logger',

  # levels
  'DEBUG',
  'INFO',
  'WARNING',
  'WARN',
  'ERROR',
  'FATAL',

  # only if you're not using app directly.
  'init',
  'teardown_stderr_logging',
  'teardown_disk_logging',

  # ditto
  'formatters',

  # other things
  'Tracer'
)



# ==================================================================================================
# Copyright 2011 Twitter, Inc.
# --------------------------------------------------------------------------------------------------
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this work except in compliance with the License.
# You may obtain a copy of the License in the LICENSE file, or at:
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==================================================================================================

"""
Interface to Glog-style formatter.

import twitter.common.log

if not using twitter.common.app:
  for __main__:
    log = twitter.common.log.init('my_binary_name')
otherwise init will be called automatically on app.main()

for library/endpoint:
  from twitter.common import log

log.info('info baby')
log.debug('debug baby')
log.fatal('oops fatal!')

Will log to my_binary_name.{INFO,WARNING,ERROR,...} into log_dir using the
Google logging format.

See twitter.com.log.options for customizations.
"""

from __future__ import print_function

import getpass
import logging
import os
from socket import gethostname
import sys
import time

from twitter.common.log.formatters import glog, plain
from twitter.common.log.handlers import ScribeHandler
from twitter.common.log.options import LogOptions
from twitter.common.dirutil import safe_mkdir


class GenericFilter(logging.Filter):
  def __init__(self, levelfn=lambda record_level: True):
    self._levelfn = levelfn
    logging.Filter.__init__(self)

  def filter(self, record):
    return 1 if self._levelfn(record.levelno) else 0


class ProxyFormatter(logging.Formatter):
  class UnknownSchemeException(Exception): pass

  _SCHEME_TO_FORMATTER = {
    glog.GlogFormatter.SCHEME: glog.GlogFormatter(),
    plain.PlainFormatter.SCHEME: plain.PlainFormatter()
  }

  def __init__(self, scheme_fn):
    logging.Formatter.__init__(self)
    self._scheme_fn = scheme_fn

  def preamble(self):
    scheme = self._scheme_fn()
    if scheme not in ProxyFormatter._SCHEME_TO_FORMATTER:
      raise ProxyFormatter.UnknownSchemeException("Unknown logging scheme: %s" % scheme)
    formatter = ProxyFormatter._SCHEME_TO_FORMATTER[scheme]
    if hasattr(formatter, 'logfile_preamble') and callable(formatter.logfile_preamble):
      return formatter.logfile_preamble()

  def format(self, record):
    scheme = self._scheme_fn()
    if scheme not in ProxyFormatter._SCHEME_TO_FORMATTER:
      raise ProxyFormatter.UnknownSchemeException("Unknown logging scheme: %s" % scheme)
    return ProxyFormatter._SCHEME_TO_FORMATTER[scheme].format(record)


_FILTER_TYPES = {
  logging.DEBUG: 'DEBUG',
  logging.INFO: 'INFO',
  logging.WARN: 'WARNING',
  logging.ERROR: 'ERROR',
  logging.FATAL: 'FATAL' # strangely python logging translates this to CRITICAL
}


def print_stderr(message):
  """Emit a message on standard error if logging to stderr is permitted."""
  if LogOptions.stderr_log_level() != LogOptions.LOG_LEVEL_NONE:
    print(message, file=sys.stderr)


def _safe_setup_link(link_filename, real_filename):
  """
    Create a symlink from link_filename to real_filename.
  """
  real_filename = os.path.relpath(real_filename, os.path.dirname(link_filename))

  if os.path.exists(link_filename):
    try:
      os.unlink(link_filename)
    except OSError:
      pass
  try:
    os.symlink(real_filename, link_filename)
  except OSError as e:
    # Typically permission denied.
    pass


class PreambleFileHandler(logging.FileHandler):
  def __init__(self, filename, preamble=None):
    self._preamble = preamble
    logging.FileHandler.__init__(self, filename)

  def _open(self):
    stream = logging.FileHandler._open(self)
    if self._preamble:
      stream.write(self._preamble)
    return stream


def _initialize_disk_logging():
  safe_mkdir(LogOptions.log_dir())


def _setup_aggregated_disk_logging(filebase):
  filename = os.path.join(LogOptions.log_dir(), filebase + '.log')
  formatter = ProxyFormatter(LogOptions.disk_log_scheme)
  file_handler = PreambleFileHandler(filename, formatter.preamble())
  file_handler.setFormatter(formatter)
  file_handler.addFilter(GenericFilter(lambda level: level >= LogOptions.disk_log_level()))
  return [file_handler]


def _setup_disk_logging(filebase):
  handlers = []

  def gen_filter(level):
    return GenericFilter(
      lambda record_level: record_level == level and level >= LogOptions.disk_log_level())

  def gen_link_filename(filebase, level):
    return '%(filebase)s.%(level)s' % {
      'filebase': filebase,
      'level': level,
    }

  hostname = gethostname()
  username = getpass.getuser()
  pid = os.getpid()
  datestring = time.strftime('%Y%m%d-%H%M%S', time.localtime())

  def gen_verbose_filename(filebase, level):
    return '%(filebase)s.%(hostname)s.%(user)s.log.%(level)s.%(date)s.%(pid)s' % {
      'filebase': filebase,
      'hostname': hostname,
      'user': username,
      'level': level,
      'date': datestring,
      'pid': pid
    }

  logroot = LogOptions.log_dir()
  for filter_type, filter_name in _FILTER_TYPES.items():
    formatter = ProxyFormatter(LogOptions.disk_log_scheme)
    filter = gen_filter(filter_type)
    full_filebase = os.path.join(logroot, filebase)
    logfile_link = gen_link_filename(full_filebase, filter_name)
    logfile_full = gen_verbose_filename(full_filebase, filter_name)
    file_handler = PreambleFileHandler(logfile_full, formatter.preamble())
    file_handler.setFormatter(formatter)
    file_handler.addFilter(filter)
    handlers.append(file_handler)
    _safe_setup_link(logfile_link, logfile_full)
  return handlers


def _setup_scribe_logging():
  filter = GenericFilter(lambda r_l: r_l >= LogOptions.scribe_log_level())
  formatter = ProxyFormatter(LogOptions.scribe_log_scheme)
  scribe_handler = ScribeHandler(buffer=LogOptions.scribe_buffer(),
                                 category=LogOptions.scribe_category(),
                                 host=LogOptions.scribe_host(),
                                 port=LogOptions.scribe_port())
  scribe_handler.setFormatter(formatter)
  scribe_handler.addFilter(filter)
  return [scribe_handler]


def _setup_stderr_logging():
  filter = GenericFilter(lambda r_l: r_l >= LogOptions.stderr_log_level())
  formatter = ProxyFormatter(LogOptions.stderr_log_scheme)
  stderr_handler = logging.StreamHandler(sys.stderr)
  stderr_handler.setFormatter(formatter)
  stderr_handler.addFilter(filter)
  return [stderr_handler]


def teardown_disk_logging():
  root_logger = logging.getLogger()
  global _DISK_LOGGERS
  for handler in _DISK_LOGGERS:
    root_logger.removeHandler(handler)
  _DISK_LOGGERS = []


def teardown_scribe_logging():
  root_logger = logging.getLogger()
  global _SCRIBE_LOGGERS
  for handler in _SCRIBE_LOGGERS:
    root_logger.removeHandler(handler)
  _SCRIBE_LOGGERS = []


def teardown_stderr_logging():
  root_logger = logging.getLogger()
  global _STDERR_LOGGERS
  for handler in _STDERR_LOGGERS:
    root_logger.removeHandler(handler)
  _STDERR_LOGGERS = []


_SCRIBE_LOGGERS = []
_STDERR_LOGGERS = []
_DISK_LOGGERS = []


def init(filebase=None):
  """
    Sets up default stderr logging and, if filebase is supplied, sets up disk logging using:
      {--log_dir}/filebase.{INFO,WARNING,...}

    If '--log_simple' is specified, logs are written into a single file:
      {--log_dir}/filebase.log
  """
  logging._acquireLock()

  # set up permissive logger
  root_logger = logging.getLogger()
  root_logger.setLevel(logging.DEBUG)

  # clear existing handlers
  teardown_scribe_logging()
  teardown_stderr_logging()
  teardown_disk_logging()
  for handler in root_logger.handlers:
    root_logger.removeHandler(handler)

  # setup INFO...FATAL handlers
  if filebase:
    _initialize_disk_logging()
    initializer = _setup_aggregated_disk_logging if LogOptions.simple() else _setup_disk_logging
    for handler in initializer(filebase):
      root_logger.addHandler(handler)
      _DISK_LOGGERS.append(handler)

  if LogOptions._is_scribe_logging_required():
    try:
      for handler in _setup_scribe_logging():
        root_logger.addHandler(handler)
        _SCRIBE_LOGGERS.append(handler)
    except ScribeHandler.ScribeHandlerException as err:
      print_stderr(err)

  for handler in _setup_stderr_logging():
    root_logger.addHandler(handler)
    _STDERR_LOGGERS.append(handler)

  logging._releaseLock()

  if len(_DISK_LOGGERS) > 0:
    print_stderr('Writing log files to disk in %s' % LogOptions.log_dir())
  if len(_SCRIBE_LOGGERS) > 0:
    print_stderr('Sending log messages to scribe host=%s:%d category=%s'
          % (LogOptions.scribe_host(), LogOptions.scribe_port(), LogOptions.scribe_category()))

  return root_logger
